
#include <viz_system.h>

class Client
{
public:
	void	*socket;

	Client	*next;
	Client	*prev;

	int	state;
	uint8_t	*input;
	uint8_t	*p;
};

static int	clientCnt;
static Client	*clients;

static void addClient(
void *s
)
{
    Client *client= new Client;
    client->socket= s;
    client->prev= 0;
    client->next= clients;
    client->state= 0;
    client->input= 0;
    client->p= 0;
    clients= client;

    ++clientCnt;
}

static void removeClient(
Client *c
)
{
    if(c->next) c->next->prev= c->prev;
    if(c->prev) c->prev->next= c->next;
    else clients= c->next;

    if(c->input) system::free(c->input);
    system::closeSocket(c->socket);
    delete c;

    --clientCnt;
}

static void clientInput(
	void	*gps,
	Client	*client,
const	uint8_t	*input,
	int	size
)
{
    if(client->input==0)
    {
	client->input= (uint8_t*)system::malloc(82);
	client->p= client->input;
	client->state= 0;
    }

    while(size--)
    {
	int c= *(input++);
	if(client->state==0 && c=='$') client->state= 1;
	if(client->state==1)
	{
	    if(c=='\r')
	    {
		*(client->p++)= '\r';
		*(client->p++)= '\n';
		client->p[0]= 0;

		const uint8_t *p= client->input;
		int n= client->p-client->input;
		while(n>0)
		{
		    int r= system::writeSerial(gps,p,n);
		    if(r>=0) n-= r;
		    else if(system::lastError()!=system::eAgain && system::lastError()!=system::eInterrupt) break;
		}

		client->p= client->input;
		client->state= 0;
	    }
	    else
	    {
		*(client->p++)= c;
		if(client->p-client->input>=80)
		{
		    client->state= 0;
		    client->p= client->input;
		}
	    }
	}
    }
}

static void broadcast(
const	uint8_t	*buffer,
	int	size
)
{
    Client *c= clients;
    while(c)
    {
	int n= size;
	Client *next= c->next;
	const uint8_t *p= buffer;
	while(n>0)
	{
	    int w= system::writeSocket(c->socket,p,n);
	    if(w>=0) n-= w;
	    else if(
		system::lastError()!=system::eAgain	&&
		system::lastError()!=system::eInterrupt
	    )
	    {
		removeClient(c);
		n= -1;
	    }
	}
	c= next;
    }
}

static inline uint32_t quickRnd(
uint32_t range
)
{
    static uint32_t state= 0x23653;
    state= (state*66049+3907864577U);
    return (uint32_t)((state/4294967295.0)*range);
}

static void shuffle(
int	nfd,
void	**sockets
)
{
    for(int i=0;i<nfd/2;++i)
    {
	int i1= quickRnd(nfd);
	int i2= quickRnd(nfd);
	void *s1= sockets[i1];
	void *s2= sockets[i2];
	sockets[i1]= s2;
	sockets[i2]= s1;
    }
}

int main(
int	,
char	**argv
)
{
    const char *portA= 0;
    const char *speedA= 0;
    const char *serial= 0;
    ++argv;

    system::error(0,"GPS3D V1.13  - gpsd - Please do not use for navigation.\n");

    while(1)
    {
	const char *arg= *(argv++);
	if(arg==0) break;

	if(system::strcmp(arg,"-port")==0)
	{
	    const char *opt= *(argv++);
	    if(opt==0) system::error(1,"-port <listen port>\n");
	    portA= opt;
	}
	else if(system::strcmp(arg,"-speed")==0)
	{
	    const char *opt= *(argv++);
	    if(opt==0) system::error(1,"-speed <serial speed>\n");
	    speedA= opt;
	}
	else if(system::strcmp(arg,"-serial")==0)
	{
	    const char *opt= *(argv++);
	    if(opt==0) system::error(1,"-serial <serial port>\n");
	    serial= opt;
	}
    }

    int port;
    int speed;
    if(serial==0) serial= system::defaultSerial();
    if(portA==0 || 1!=system::sscanf(portA,"%d",&port)) port= 2222;
    if(speedA==0 || 1!=system::sscanf(speedA,"%d",&speed)) speed= 19200;

    void *gps= system::openSerial(serial,speed);
    if(gps==(void*)-1) system::error(1,"fatal: Couldn't open serial port %s\n",serial);

    const char *mask= "0.0.0.0";
    void *lsock= system::makeListenSocket(mask,port);
    if(lsock==(void*)-1) system::error(1,"fatal: Couldn't create listening socket on port %s:%d\n",mask,port);

    system::daemonize();
    system::ignoreSignals();

    void **fds= 0;
    int fdsAlloc= 0;
    while(1)
    {
	int nfd= 2+clientCnt;
	if(nfd>fdsAlloc)
	{
	    fdsAlloc= 2*nfd+10;
	    fds= (void**)system::realloc(fds,fdsAlloc*sizeof(void*));
	}

	void **fp= fds;
	*(fp++)= gps;
	*(fp++)= lsock;

	Client *c= clients;
	while(c)
	{
	    *(fp++)= c->socket;
	    c= c->next;
	}

	shuffle(nfd,fds);

	void *r= system::waitOn(nfd,fds);
	if(r==gps)
	{
	    while(1)
	    {
		uint8_t buffer[2048];
		int r= system::readSerial(gps,buffer,2048);
		if(r<=0) break;
		broadcast(buffer,r);
	    }
	}
	else if(r==lsock)
	{
	    void *asock= system::acceptConnection(lsock);
	    if(asock!=(void*)-1) addClient(asock);
	}
	else
	{
	    c= clients;
	    while(c)
	    {
		void *s= c->socket;
		Client *next= c->next;
		if(s==r)
		{
		    uint8_t buffer[2048];
		    int r= system::readSocket(s,buffer,2048);
		    if(r>0) clientInput(gps,c,buffer,r);
		    else if(system::lastError()==system::eAgain) break;
		    else if(system::lastError()!=system::eInterrupt)
		    {
			removeClient(c);
			break;
		    }
		}
		c= next;
	    }
	}
    }
    return 0;
}

